import pytest
from saltfactories.utils.functional import StateResult

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.mark.slow_test
def test_echo(salt_ssh_cli, base_env_state_tree_root_dir):
    """
    verify salt-ssh can use imported map files in states
    """
    name = "echo"
    echo = "hello"
    state_file = """
    ssh_test_echo:
      test.show_notification:
        - text: {{{{ salt['test.echo']('{echo}') }}}}
    """.format(
        echo=echo
    )
    state_tempfile = pytest.helpers.temp_file(
        f"{name}.sls", state_file, base_env_state_tree_root_dir
    )

    with state_tempfile:
        ret = salt_ssh_cli.run("state.apply", name)
        result = StateResult(ret.data)
        assert result.comment == echo



"""
    salt._logging.handlers
    ~~~~~~~~~~~~~~~~~~~~~~

    Salt's logging handlers
"""

import logging
import logging.handlers
import sys
from collections import deque

from salt._logging.mixins import ExcInfoOnLogLevelFormatMixin

log = logging.getLogger(__name__)


class StreamHandler(ExcInfoOnLogLevelFormatMixin, logging.StreamHandler):
    """
    Stream handler which properly handles exc_info on a per handler basis
    """


class DeferredStreamHandler(StreamHandler):
    """
    This logging handler will store all the log records up to its maximum
    queue size at which stage the first messages stored will be dropped.

    Should only be used as a temporary logging handler, while the logging
    system is not fully configured.

    Once configured, pass any logging handlers that should have received the
    initial log messages to the function
    :func:`DeferredStreamHandler.sync_with_handlers` and all stored log
    records will be dispatched to the provided handlers.

    If anything goes wrong before logging is properly setup, all stored messages
    will be flushed to the handler's stream, ie, written to console.

    .. versionadded:: 3005
    """

    def __init__(self, stream, max_queue_size=10000):
        super().__init__(stream)
        self.__messages = deque(maxlen=max_queue_size)
        self.__emitting = False

    def handle(self, record):
        self.acquire()
        self.__messages.append(record)
        self.release()

    def flush(self):
        if self.__emitting:
            # We set the flushing flag because the stream handler actually calls flush when
            # emitting a log record and we don't want to cause a RecursionError
            return
        while self.__messages:
            try:
                self.__emitting = True
                record = self.__messages.popleft()
                # We call the parent's class handle method so it's actually
                # handled and not queued back
                # However, temporarily
                super().handle(record)
            finally:
                self.__emitting = False
        # Seeing an exception from calling flush on a closed file in the test
        # suite. Handling this condition for now but this seems to be
        # indicitive of an un-clean teardown at some point.
        if not self.stream.closed:
            super().flush()

    def sync_with_handlers(self, handlers=()):
        """
        Sync the stored log records to the provided log handlers.
        """
        while self.__messages:
            record = self.__messages.popleft()
            for handler in handlers:
                if handler is self:
                    continue
                handler.handle(record)


class FileHandler(ExcInfoOnLogLevelFormatMixin, logging.FileHandler):
    """
    File handler which properly handles exc_info on a per handler basis
    """


class SysLogHandler(ExcInfoOnLogLevelFormatMixin, logging.handlers.SysLogHandler):
    """
    Syslog handler which properly handles exc_info on a per handler basis
    """

    def handleError(self, record):
        """
        Override the default error handling mechanism for py3
        Deal with syslog os errors when the log file does not exist
        """
        handled = False
        if sys.stderr and sys.version_info >= (3, 5, 4):
            exc_type, exc, exc_traceback = sys.exc_info()
            try:
                if exc_type.__name__ in "FileNotFoundError":
                    sys.stderr.write(
                        "[WARNING ] The log_file does not exist. Logging not "
                        "setup correctly or syslog service not started.\n"
                    )
                    handled = True
            finally:
                # 'del' recommended. See documentation of
                # 'sys.exc_info()' for details.
                del exc_type, exc, exc_traceback

        if not handled:
            super().handleError(record)


class RotatingFileHandler(
    ExcInfoOnLogLevelFormatMixin, logging.handlers.RotatingFileHandler
):
    """
    Rotating file handler which properly handles exc_info on a per handler basis
    """

    def handleError(self, record):
        """
        Override the default error handling mechanism

        Deal with log file rotation errors due to log file in use
        more softly.
        """
        handled = False

        # Can't use "salt.utils.platform.is_windows()" in this file
        if (
            sys.platform.startswith("win") and logging.raiseExceptions and sys.stderr
        ):  # see Python issue 13807
            exc_type, exc, exc_traceback = sys.exc_info()
            try:
                # PermissionError is used since Python 3.3.
                # OSError is used for previous versions of Python.
                if (
                    exc_type.__name__ in ("PermissionError", "OSError")
                    and exc.winerror == 32
                ):
                    if self.level <= logging.WARNING:
                        sys.stderr.write(
                            '[WARNING ] Unable to rotate the log file "{}" '
                            "because it is in use\n".format(self.baseFilename)
                        )
                    handled = True
            finally:
                # 'del' recommended. See documentation of
                # 'sys.exc_info()' for details.
                del exc_type, exc, exc_traceback

        if not handled:
            super().handleError(record)


class WatchedFileHandler(
    ExcInfoOnLogLevelFormatMixin, logging.handlers.WatchedFileHandler
):
    """
    Watched file handler which properly handles exc_info on a per handler basis
    """



"""
    :codeauthor: Thomas Jackson <jacksontj.89@gmail.com>
"""

import logging
import threading

import pytest
from pytestshellutils.utils import ports

import salt.channel.client
import salt.channel.server
import salt.config
import salt.exceptions
import salt.ext.tornado.concurrent
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.utils.platform
import salt.utils.process
from salt.ext.tornado.testing import AsyncTestCase
from tests.support.mixins import AdaptedConfigurationTestCaseMixin
from tests.unit.transport.mixins import run_loop_in_thread

pytestmark = [
    pytest.mark.skip_on_darwin,
    pytest.mark.skip_on_freebsd,
]

log = logging.getLogger(__name__)


@pytest.mark.skip(reason="Skip until we can devote time to fix this test")
class AsyncPubServerTest(AsyncTestCase, AdaptedConfigurationTestCaseMixin):
    """
    Tests around the publish system
    """

    @classmethod
    def setUpClass(cls):
        ret_port = ports.get_unused_localhost_port()
        publish_port = ports.get_unused_localhost_port()
        tcp_master_pub_port = ports.get_unused_localhost_port()
        tcp_master_pull_port = ports.get_unused_localhost_port()
        tcp_master_publish_pull = ports.get_unused_localhost_port()
        tcp_master_workers = ports.get_unused_localhost_port()
        cls.master_config = cls.get_temp_config(
            "master",
            **{
                "transport": "tcp",
                "auto_accept": True,
                "ret_port": ret_port,
                "publish_port": publish_port,
                "tcp_master_pub_port": tcp_master_pub_port,
                "tcp_master_pull_port": tcp_master_pull_port,
                "tcp_master_publish_pull": tcp_master_publish_pull,
                "tcp_master_workers": tcp_master_workers,
            },
        )

        cls.minion_config = cls.get_temp_config(
            "minion",
            **{
                "transport": "tcp",
                "master_ip": "127.0.0.1",
                "auth_timeout": 1,
                "master_port": ret_port,
                "master_uri": f"tcp://127.0.0.1:{ret_port}",
            },
        )

        cls.process_manager = salt.utils.process.ProcessManager(
            name="ReqServer_ProcessManager"
        )

        cls.server_channel = salt.channel.server.PubServerChannel.factory(
            cls.master_config
        )
        cls.server_channel.pre_fork(cls.process_manager)

        # we also require req server for auth
        cls.req_server_channel = salt.channel.server.ReqServerChannel.factory(
            cls.master_config
        )
        cls.req_server_channel.pre_fork(cls.process_manager)
        cls.io_loop = salt.ext.tornado.ioloop.IOLoop()
        cls.stop = threading.Event()
        cls.req_server_channel.post_fork(cls._handle_payload, io_loop=cls.io_loop)
        cls.server_thread = threading.Thread(
            target=run_loop_in_thread,
            args=(
                cls.io_loop,
                cls.stop,
            ),
        )
        cls.server_thread.start()

    @classmethod
    def _handle_payload(cls, payload):
        """
        TODO: something besides echo
        """
        return payload, {"fun": "send_clear"}

    @classmethod
    def tearDownClass(cls):
        cls.req_server_channel.close()
        cls.server_channel.close()
        cls.stop.set()
        cls.server_thread.join()
        cls.process_manager.terminate()
        del cls.req_server_channel

    def setUp(self):
        super().setUp()
        self._start_handlers = dict(self.io_loop._handlers)

    def tearDown(self):
        super().tearDown()
        failures = []
        for k, v in self.io_loop._handlers.items():
            if self._start_handlers.get(k) != v:
                failures.append((k, v))
        if failures:
            raise Exception(f"FDs still attached to the IOLoop: {failures}")
        del self.channel
        del self._start_handlers

    def test_basic(self):
        self.pub = None

        def handle_pub(ret):
            self.pub = ret
            self.stop()  # pylint: disable=not-callable

        self.pub_channel = salt.channel.client.AsyncPubChannel.factory(
            self.minion_opts, io_loop=self.io_loop
        )
        connect_future = self.pub_channel.connect()
        connect_future.add_done_callback(
            lambda f: self.stop()  # pylint: disable=not-callable
        )
        self.wait()
        connect_future.result()
        self.pub_channel.on_recv(handle_pub)
        load = {
            "fun": "f",
            "arg": "a",
            "tgt": "t",
            "jid": "j",
            "ret": "r",
            "tgt_type": "glob",
        }
        self.server_channel.publish(load)
        self.wait()
        self.assertEqual(self.pub["load"], load)
        self.pub_channel.on_recv(None)
        self.server_channel.publish(load)
        with self.assertRaises(self.failureException):
            self.wait(timeout=0.5)

        # close our pub_channel, to pass our FD checks
        self.pub_channel.close()
        del self.pub_channel



"""
This module is a central location for all salt exceptions
"""

import copy
import logging
import time

import salt.defaults.exitcodes

log = logging.getLogger(__name__)


def _nested_output(obj):
    """
    Serialize obj and format for output
    """
    # Explicit late import to avoid circular import
    from salt.output import nested

    nested.__opts__ = {}
    ret = nested.output(obj).rstrip()
    return ret


def get_error_message(error):
    """
    Get human readable message from Python Exception
    """
    return error.args[0] if error.args else ""


class SaltException(Exception):
    """
    Base exception class; all Salt-specific exceptions should subclass this
    """

    def __init__(self, message=""):
        # Avoid circular import
        import salt.utils.stringutils

        if not isinstance(message, str):
            message = str(message)
        super().__init__(salt.utils.stringutils.to_str(message))
        self.message = self.strerror = message

    def pack(self):
        """
        Pack this exception into a serializable dictionary that is safe for
        transport via msgpack
        """
        return {"message": str(self), "args": self.args}


class SaltClientError(SaltException):
    """
    Problem reading the master root key
    """


class SaltMasterError(SaltException):
    """
    Problem reading the master root key
    """


class SaltNoMinionsFound(SaltException):
    """
    An attempt to retrieve a list of minions failed
    """


class SaltSyndicMasterError(SaltException):
    """
    Problem while proxying a request in the syndication master
    """


class SaltMasterUnresolvableError(SaltException):
    """
    Problem resolving the name of the Salt master
    """


class MasterExit(SystemExit):
    """
    Rise when the master exits
    """


class AuthenticationError(SaltException):
    """
    If sha256 signature fails during decryption
    """


class InvalidKeyError(SaltException):
    """
    Raised when we encounter an invalid RSA key.
    """


class CommandNotFoundError(SaltException):
    """
    Used in modules or grains when a required binary is not available
    """


class CommandExecutionError(SaltException):
    """
    Used when a module runs a command which returns an error and wants
    to show the user the output gracefully instead of dying
    """

    def __init__(self, message="", info=None):
        # Avoid circular import
        import salt.utils.stringutils

        try:
            exc_str_prefix = salt.utils.stringutils.to_unicode(message)
        except TypeError:
            # Exception class instance passed. The SaltException __init__ will
            # gracefully handle non-string types passed to it, but since this
            # class needs to do some extra stuff with the exception "message"
            # before handing it off to the parent class' __init__, we'll need
            # to extract the message from the exception instance here
            try:
                exc_str_prefix = str(message)
            except UnicodeDecodeError:
                exc_str_prefix = salt.utils.stringutils.to_unicode(str(message))
        self.error = exc_str_prefix
        self.info = info
        if self.info:
            if exc_str_prefix:
                if exc_str_prefix[-1] not in ".?!":
                    exc_str_prefix += "."
                exc_str_prefix += " "

            exc_str_prefix += "Additional info follows:\n\n"
            # NOTE: exc_str will be passed to the parent class' constructor and
            # become self.strerror.
            exc_str = exc_str_prefix + _nested_output(self.info)

            # For states, if self.info is a dict also provide an attribute
            # containing a nested output of the info dict without the changes
            # (since they will be in the 'changes' key of the state return and
            # this information would be redundant).
            if isinstance(self.info, dict):
                info_without_changes = copy.deepcopy(self.info)
                info_without_changes.pop("changes", None)
                if info_without_changes:
                    self.strerror_without_changes = exc_str_prefix + _nested_output(
                        info_without_changes
                    )
                else:
                    # 'changes' was the only key in the info dictionary. We no
                    # longer have any additional info to display. Use the
                    # original error message.
                    self.strerror_without_changes = self.error
            else:
                self.strerror_without_changes = exc_str
        else:
            self.strerror_without_changes = exc_str = self.error

        # We call the parent __init__ last instead of first because we need the
        # logic above to derive the message string to use for the exception
        # message.
        super().__init__(exc_str)


class LoaderError(SaltException):
    """
    Problems loading the right renderer
    """


class PublishError(SaltException):
    """
    Problems encountered when trying to publish a command
    """


class MinionError(SaltException):
    """
    Minion problems reading uris such as salt:// or http://
    """


class FileserverConfigError(SaltException):
    """
    Used when invalid fileserver settings are detected
    """


class FileLockError(SaltException):
    """
    Used when an error occurs obtaining a file lock
    """

    def __init__(
        self, message, time_start=None, *args, **kwargs
    ):  # pylint: disable=keyword-arg-before-vararg
        super().__init__(message, *args, **kwargs)
        if time_start is None:
            log.warning(
                "time_start should be provided when raising a FileLockError. "
                "Defaulting to current time as a fallback, but this may "
                "result in an inaccurate timeout."
            )
            self.time_start = time.time()
        else:
            self.time_start = time_start


class GitLockError(SaltException):
    """
    Raised when an uncaught error occurs in the midst of obtaining an
    update/checkout lock in salt.utils.gitfs.

    NOTE: While this uses the errno param similar to an OSError, this exception
    class is *not* as subclass of OSError. This is done intentionally, so that
    this exception class can be caught in a try/except without being caught as
    an OSError.
    """

    def __init__(self, errno, message, *args, **kwargs):
        super().__init__(message, *args, **kwargs)
        self.errno = errno


class GitRemoteError(SaltException):
    """
    Used by GitFS to denote a problem with the existence of the "origin" remote
    or part of its configuration
    """


class SaltInvocationError(SaltException, TypeError):
    """
    Used when the wrong number of arguments are sent to modules or invalid
    arguments are specified on the command line
    """


class PkgParseError(SaltException):
    """
    Used when of the pkg modules cannot correctly parse the output from
    the CLI tool (pacman, yum, apt, aptitude, etc)
    """


class SaltRenderError(SaltException):
    """
    Used when a renderer needs to raise an explicit error. If a line number and
    buffer string are passed, get_context will be invoked to get the location
    of the error.
    """

    def __init__(
        self,
        message,
        line_num=None,
        buf="",
        marker="    <======================",
        trace=None,
    ):
        # Avoid circular import
        import salt.utils.stringutils

        self.error = message
        try:
            exc_str = salt.utils.stringutils.to_unicode(message)
        except TypeError:
            # Exception class instance passed. The SaltException __init__ will
            # gracefully handle non-string types passed to it, but since this
            # class needs to do some extra stuff with the exception "message"
            # before handing it off to the parent class' __init__, we'll need
            # to extract the message from the exception instance here
            try:
                exc_str = str(message)
            except UnicodeDecodeError:
                exc_str = salt.utils.stringutils.to_unicode(str(message))
        self.line_num = line_num
        self.buffer = buf
        self.context = ""
        if trace:
            exc_str += f"\n{trace}\n"
        if self.line_num and self.buffer:
            # Avoid circular import
            import salt.utils.templates

            self.context = salt.utils.stringutils.get_context(
                self.buffer, self.line_num, marker=marker
            )
            exc_str += "; line {}\n\n{}".format(
                self.line_num,
                salt.utils.stringutils.to_unicode(self.context),
            )
        super().__init__(exc_str)


class SaltClientTimeout(SaltException):
    """
    Thrown when a job sent through one of the Client interfaces times out

    Takes the ``jid`` as a parameter
    """

    def __init__(
        self, message, jid=None, *args, **kwargs
    ):  # pylint: disable=keyword-arg-before-vararg
        super().__init__(message, *args, **kwargs)
        self.jid = jid


class SaltCacheError(SaltException):
    """
    Thrown when a problem was encountered trying to read or write from the salt cache
    """


class TimeoutError(SaltException):
    """
    Thrown when an opration cannot be completet within a given time limit.
    """


class SaltReqTimeoutError(SaltException):
    """
    Thrown when a salt master request call fails to return within the timeout
    """


class TimedProcTimeoutError(SaltException):
    """
    Thrown when a timed subprocess does not terminate within the timeout,
    or if the specified timeout is not an int or a float
    """


class EauthAuthenticationError(SaltException):
    """
    Thrown when eauth authentication fails
    """


class TokenAuthenticationError(SaltException):
    """
    Thrown when token authentication fails
    """


class SaltDeserializationError(SaltException):
    """
    Thrown when salt cannot deserialize data.
    """


class AuthorizationError(SaltException):
    """
    Thrown when runner or wheel execution fails due to permissions
    """


class SaltValidationError(SaltException):
    """
    Thrown when a value fails validation
    """


class UnsupportedAlgorithm(SaltException):
    """
    Thrown when a requested encryption or signing algorithm is un-supported.
    """


class SaltDaemonNotRunning(SaltException):
    """
    Throw when a running master/minion/syndic is not running but is needed to
    perform the requested operation (e.g., eauth).
    """


class SaltRunnerError(SaltException):
    """
    Problem in runner
    """


class SaltWheelError(SaltException):
    """
    Problem in wheel
    """


class SaltConfigurationError(SaltException):
    """
    Configuration error
    """


class SaltSystemExit(SystemExit):
    """
    This exception is raised when an unsolvable problem is found. There's
    nothing else to do, salt should just exit.
    """

    def __init__(self, code=0, msg=None):
        SystemExit.__init__(self, msg)


class SaltCloudException(SaltException):
    """
    Generic Salt Cloud Exception
    """


class SaltCloudSystemExit(SaltCloudException):
    """
    This exception is raised when the execution should be stopped.
    """

    def __init__(self, message, exit_code=salt.defaults.exitcodes.EX_GENERIC):
        super().__init__(message)
        self.message = message
        self.exit_code = exit_code


class SaltCloudConfigError(SaltCloudException):
    """
    Raised when a configuration setting is not found and should exist.
    """


class SaltCloudNotFound(SaltCloudException):
    """
    Raised when some cloud provider function cannot find what's being searched.
    """


class SaltCloudExecutionTimeout(SaltCloudException):
    """
    Raised when too much time has passed while querying/waiting for data.
    """


class SaltCloudExecutionFailure(SaltCloudException):
    """
    Raised when too much failures have occurred while querying/waiting for data.
    """


class SaltCloudPasswordError(SaltCloudException):
    """
    Raise when virtual terminal password input failed
    """


class NotImplemented(SaltException):
    """
    Used when a module runs a command which returns an error and wants
    to show the user the output gracefully instead of dying
    """


class TemplateError(SaltException):
    """
    Used when a custom error is triggered in a template
    """


class ArgumentValueError(CommandExecutionError):
    """
    Used when an invalid argument was passed to a command execution
    """


class CheckError(CommandExecutionError):
    """
    Used when a check fails
    """


# Validation related exceptions
class InvalidConfigError(CommandExecutionError):
    """
    Used when the config is invalid
    """


class InvalidEntityError(CommandExecutionError):
    """
    Used when an entity fails validation
    """


# VMware related exceptions
class VMwareSaltError(CommandExecutionError):
    """
    Used when a VMware object cannot be retrieved
    """


class VMwareRuntimeError(VMwareSaltError):
    """
    Used when a runtime error is encountered when communicating with the
    vCenter
    """


class VMwareConnectionError(VMwareSaltError):
    """
    Used when the client fails to connect to a either a VMware vCenter server or
    to a ESXi host
    """


class VMwareObjectRetrievalError(VMwareSaltError):
    """
    Used when a VMware object cannot be retrieved
    """


class VMwareObjectNotFoundError(VMwareSaltError):
    """
    Used when a VMware object was not found
    """


class VMwareObjectExistsError(VMwareSaltError):
    """
    Used when a VMware object already exists
    """


class VMwareMultipleObjectsError(VMwareObjectRetrievalError):
    """
    Used when multiple objects were retrieved (and one was expected)
    """


class VMwareNotFoundError(VMwareSaltError):
    """
    Used when a VMware object was not found
    """


class VMwareApiError(VMwareSaltError):
    """
    Used when representing a generic VMware API error
    """


class VMwareFileNotFoundError(VMwareApiError):
    """
    Used when representing a generic VMware error if a file not found
    """


class VMwareSystemError(VMwareSaltError):
    """
    Used when representing a generic VMware system error
    """


class VMwarePowerOnError(VMwareSaltError):
    """
    Used when error occurred during power on
    """


class VMwareVmRegisterError(VMwareSaltError):
    """
    Used when a configuration parameter is incorrect
    """


class VMwareVmCreationError(VMwareSaltError):
    """
    Used when a configuration parameter is incorrect
    """


class MissingSmb(SaltException):
    """
    Raised when no smb library is found.
    """


class LoggingRuntimeError(RuntimeError):
    """
    Raised when we encounter an error while logging
    """


class NxosError(SaltException):
    """
    NX-OS Base Exception class
    """


class NxosCliError(NxosError):
    """
    NX-OS Cli Error raised when Cli command rejected by the NX-OS device
    """


class NxosClientError(NxosError):
    """
    NX-OS Client Error raised for problems connecting to the NX-OS device
    """


class NxosRequestNotSupported(NxosError):
    """
    Raised for unsupported client requests
    """


class CodePageError(CommandExecutionError):
    """
    Raised when an error ocurs while getting or setting the windows code page
    """



import tempfile

import salt.modules.ssh as ssh
import salt.utils.files
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, mock_open, patch
from tests.support.unit import TestCase


class SSHAuthKeyTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.ssh
    """

    def setup_loader_modules(self):
        return {
            ssh: {
                "__salt__": {
                    "user.info": lambda u: getattr(self, "user_info_mock", None),
                }
            }
        }

    def tearDown(self):
        try:
            delattr(self, "user_info_mock")
        except AttributeError:
            pass

    def test_expand_user_token(self):
        """
        Test if the %u, %h, and %% tokens are correctly expanded
        """
        output = ssh._expand_authorized_keys_path("/home/%u", "user", "/home/user")
        self.assertEqual(output, "/home/user")

        output = ssh._expand_authorized_keys_path("/home/%h", "user", "/home/user")
        self.assertEqual(output, "/home//home/user")

        output = ssh._expand_authorized_keys_path("%h/foo", "user", "/home/user")
        self.assertEqual(output, "/home/user/foo")

        output = ssh._expand_authorized_keys_path(
            "/srv/%h/aaa/%u%%", "user", "/home/user"
        )
        self.assertEqual(output, "/srv//home/user/aaa/user%")

        user = "dude"
        home = "/home/dude"
        path = "/home/dude%"
        self.assertRaises(
            CommandExecutionError, ssh._expand_authorized_keys_path, path, user, home
        )

        path = "/home/%dude"
        self.assertRaises(
            CommandExecutionError, ssh._expand_authorized_keys_path, path, user, home
        )

    def test_set_auth_key_invalid(self):
        self.user_info_mock = {"home": "/dev/null"}
        # Inserting invalid public key should be rejected
        invalid_key = "AAAAB3NzaC1kc3MAAACBAL0sQ9fJ5bYTEyY"  # missing padding
        self.assertEqual(ssh.set_auth_key("user", invalid_key), "Invalid public key")

    def test_set_auth_key_sk_ed25519_cert(self):
        self.user_info_mock = {"home": "/dev/null"}
        valid_key = (
            "AAAAI3NrLXNzaC1lZDI1NTE5LWNlcnQtdjAxQG9wZW5zc2guY29tAAAAIPMSTMKu"
            "DdipIQl8IA3UXl5WHYcIyF2tfwrri/Wd/oV3AAAAIKuzsywoer6Y7oYtLXse/TKj"
            "VqqKjEpUq+4zMkQ9FEwJAAAABHNzaDoAAAAAAAAAAAAAAAEAAAARbWF4QHNjaHUg"
            "dXNlciBrZXkAAAAHAAAAA21heAAAAABgGcIwAAAAAHLlxlwAAAAAAAAAggAAABVw"
            "ZXJtaXQtWDExLWZvcndhcmRpbmcAAAAAAAAAF3Blcm1pdC1hZ2VudC1mb3J3YXJk"
            "aW5nAAAAAAAAABZwZXJtaXQtcG9ydC1mb3J3YXJkaW5nAAAAAAAAAApwZXJtaXQt"
            "cHR5AAAAAAAAAA5wZXJtaXQtdXNlci1yYwAAAAAAAAAAAAAAMwAAAAtzc2gtZWQy"
            "NTUxOQAAACDI8h3vsne8ZtyH7JRHmkImHXQciefsH4e99ka3HJzKPQAAAFMAAAAL"
            "c3NoLWVkMjU1MTkAAABAHzruY+hPXK2ONt9d2XFUttvdSR7dW9Yy7stru4zopgUb"
            "o0CjTlKyogb7PiRryt5JExT1Wkux1q3oBtyvSdG+Cw=="
        )
        # we expect 'fail', but not 'Invalid public key'
        self.assertEqual(
            ssh.set_auth_key(
                "user", valid_key, enc="sk-ssh-ed25519-cert-v01@openssh.com"
            ),
            "fail",
        )

    def test_set_auth_key_from_file(self):
        """
        Test if set_auth_key_from_file pass the correct options to set_auth_key and
        if multiple keys are supported.
        """
        s_keys_dict = {
            "S1KeyWithSpecialCharactersAndMultipleWordCommentaryXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "sk-ssh-ed25519@openssh.com",
                "comment": "user@internal.example Multiple Words Commentary",
                "options": [],
                "fingerprint": "c5:6c:6d:47:6e:d6:16:56:60:7b:f6:ab:59:c6:33:03:87:b8:3b:d7:67:7e:37:8a:db:83:8b:b5:48:de:91:c1",
            },
            "S4KeyWithSpecialCharactersAndMultipleWordCommentaryAndOptionsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "sk-ssh-ed25519@openssh.com",
                "comment": "user@internal.example Multiple Words Commentary",
                "options": ["no-port-forwarding", "no-pty"],
                "fingerprint": "3a:14:23:b6:db:0a:d7:07:30:30:3b:b4:9a:31:0c:29:9b:c5:fa:5f:5e:5a:7d:fb:4b:44:da:70:02:79:ca:04",
            },
        }
        user = "user"
        config = "authorized_keys_somewhere"
        options = ["no-pty"]
        fingerprint_hash_type = None

        expected_calls_without_options = []
        expected_calls_with_options = []
        for key in s_keys_dict:
            expected_calls_without_options.append(
                call(
                    user,
                    key,
                    enc=s_keys_dict[key]["enc"],
                    comment=s_keys_dict[key]["comment"],
                    options=s_keys_dict[key]["options"],
                    config=config,
                    cache_keys=list(s_keys_dict.keys()),
                    fingerprint_hash_type=fingerprint_hash_type,
                )
            )
            expected_calls_with_options.append(
                call(
                    user,
                    key,
                    enc=s_keys_dict[key]["enc"],
                    comment=s_keys_dict[key]["comment"],
                    options=options,
                    config=config,
                    cache_keys=list(s_keys_dict.keys()),
                    fingerprint_hash_type=fingerprint_hash_type,
                )
            )

        with patch.object(ssh, "set_auth_key", return_value="new") as mock_set_auth_key:
            with patch("os.path.isfile", return_value=True):
                with patch.dict(
                    ssh.__salt__,
                    {"cp.cache_file": MagicMock(return_value="/fakecache/fakefile")},
                ):
                    with patch(
                        "salt.modules.ssh._validate_keys",
                        MagicMock(return_value=s_keys_dict),
                    ):
                        self.assertEqual(
                            ssh.set_auth_key_from_file(
                                "user", source="salt://fakefile", config=config
                            ),
                            "new",
                        )
                        mock_set_auth_key.assert_has_calls(
                            expected_calls_without_options
                        )

                        self.assertEqual(
                            ssh.set_auth_key_from_file(
                                "user",
                                source="salt://fakefile",
                                config=config,
                                options=options,
                            ),
                            "new",
                        )
                        mock_set_auth_key.assert_has_calls(expected_calls_with_options)

    def test_rm_auth_key_from_file(self):
        """
        Test if multiple keys are supported by rm_auth_key_from_file
        """
        s_keys_dict = {
            "S1KeyWithSpecialCharactersAndMultipleWordCommentaryXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "sk-ssh-ed25519@openssh.com",
                "comment": "user@internal.example Multiple Words Commentary",
                "options": [],
                "fingerprint": "c5:6c:6d:47:6e:d6:16:56:60:7b:f6:ab:59:c6:33:03:87:b8:3b:d7:67:7e:37:8a:db:83:8b:b5:48:de:91:c1",
            },
            "S4KeyWithSpecialCharactersAndMultipleWordCommentaryAndOptionsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "sk-ssh-ed25519@openssh.com",
                "comment": "user@internal.example Multiple Words Commentary",
                "options": ["no-port-forwarding", "no-pty"],
                "fingerprint": "3a:14:23:b6:db:0a:d7:07:30:30:3b:b4:9a:31:0c:29:9b:c5:fa:5f:5e:5a:7d:fb:4b:44:da:70:02:79:ca:04",
            },
        }
        user = "user"
        config = "authorized_keys_somewhere"
        fingerprint_hash_type = None

        expected_calls = []
        for key in s_keys_dict:
            expected_calls.append(
                call(
                    user,
                    key,
                    config=config,
                    fingerprint_hash_type=fingerprint_hash_type,
                )
            )

        with patch.object(
            ssh, "rm_auth_key", return_value="Key removed"
        ) as mock_rm_auth_key:
            with patch("os.path.isfile", return_value=True):
                with patch.dict(
                    ssh.__salt__,
                    {"cp.cache_file": MagicMock(return_value="/fakecache/fakefile")},
                ):
                    with patch(
                        "salt.modules.ssh._validate_keys",
                        MagicMock(return_value=s_keys_dict),
                    ):
                        self.assertEqual(
                            ssh.rm_auth_key_from_file(
                                "user", source="salt://fakefile", config=config
                            ),
                            "Key removed",
                        )
                        mock_rm_auth_key.assert_has_calls(expected_calls)

    def test__validate_keys(self):
        """
        Test if keys read from file are correctly validated
        """
        s_keys_dict = {
            "KeyWithSpecialCharactersAndOptionsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "sk-ecdsa-sha2-nistp256@openssh.com",
                "comment": "user@internal.example",
                "options": ["no-port-forwarding", "no-pty"],
                "fingerprint": "b4:98:a7:02:bf:0d:1f:0a:c2:96:19:4c:10:7e:7b:51:ad:d7:ee:2c:3e:96:37:58:4d:ab:73:98:db:ee:36:f1",
            },
            "KeyWithoutOptionsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "ecdsa-sha2-nistp256",
                "comment": "user@internal.example",
                "options": [],
                "fingerprint": "4e:1f:ba:e6:16:31:56:67:f6:c1:ea:34:0c:b4:83:43:28:2c:5f:e7:c1:95:5c:07:90:6f:70:d5:4d:1b:3e:af",
            },
            "KeyWithoutOptionsAndWithoutCommentsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "ecdsa-sha2-nistp384",
                "comment": "",
                "options": [],
                "fingerprint": "3d:55:90:4b:d5:7d:22:03:37:c7:8c:4f:32:8b:5e:a9:77:f8:fd:f7:e4:24:52:87:14:90:b3:d6:c3:d8:57:f0",
            },
            "KeyWithOptionsAndWithoutCommentsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "ecdsa-sha2-nistp521",
                "comment": "",
                "options": ["no-port-forwarding", "no-pty"],
                "fingerprint": "a9:7d:6d:00:30:df:28:75:1b:1f:f1:b5:b6:36:3c:82:e4:bf:70:97:bf:6c:a6:e0:9a:b6:37:6b:0d:30:57:e8",
            },
            "KeyWithSpecialCharactersWithoutOptionsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "sk-ssh-ed25519@openssh.com",
                "comment": "user@internal.example",
                "options": [],
                "fingerprint": "3e:8b:af:7c:16:12:c6:e2:df:2d:03:16:b7:19:28:b8:3e:65:de:b0:7d:fc:74:06:a9:d6:b1:0f:92:df:7c:cf",
            },
            "KeyWithMultipleWordCommentaryAndNoOptionsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "ssh-ed25519",
                "comment": "user@internal.example Multiple Word Commentary",
                "options": [],
                "fingerprint": "29:e8:ef:0d:f4:93:09:13:19:af:66:bf:d0:ca:45:0a:ed:ac:30:e6:3b:41:22:3a:d6:ea:e6:a3:5f:8b:af:07",
            },
            "KeyWithMultipleWordCommentaryAndWithOptionsXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX=": {
                "enc": "ssh-rsa",
                "comment": "user@internal.example Multiple Word Commentary",
                "options": ["no-port-forwarding", "no-pty"],
                "fingerprint": "c7:e2:fb:76:03:ab:f0:11:2d:62:11:33:25:57:1a:1f:88:2d:cc:4f:95:44:bd:2c:79:11:8f:84:d9:75:b5:bf",
            },
        }

        read_file_contents = ""
        for key in s_keys_dict:
            line = f"{s_keys_dict[key]['enc']} {key}"
            if s_keys_dict[key]["options"]:
                options = ",".join(s_keys_dict[key]["options"])
                line = " ".join([options, line])
            if s_keys_dict[key]["comment"]:
                line = " ".join([line, s_keys_dict[key]["comment"]])
            line += "\n"
            read_file_contents += line

        fopen_mock = mock_open(read_data=read_file_contents)
        with patch("salt.utils.files.fopen", fopen_mock):
            assert ssh._validate_keys("keyfile", None) == s_keys_dict

    def test_replace_auth_key(self):
        """
        Test the _replace_auth_key with some different authorized_keys examples
        """
        # First test a known working example, gathered from the authorized_keys file
        # in the integration test files.
        enc = "ssh-rsa"
        key = (
            "AAAAB3NzaC1yc2EAAAABIwAAAQEAq2A7hRGmdnm9tUDbO9IDSwBK6TbQa+"
            "PXYPCPy6rbTrTtw7PHkccKrpp0yVhp5HdEIcKr6pLlVDBfOLX9QUsyCOV0wzfjIJNl"
            "GEYsdlLJizHhbn2mUjvSAHQqZETYP81eFzLQNnPHt4EVVUh7VfDESU84KezmD5QlWp"
            "XLmvU31/yMf+Se8xhHTvKSCZIFImWwoG6mbUoWf9nzpIoaSjB+weqqUUmpaaasXVal"
            "72J+UX2B+2RPW3RcT0eOzQgqlJL3RKrTJvdsjE3JEAvGq3lGHSZXy28G3skua2SmVi"
            "/w4yCE6gbODqnTWlg7+wC604ydGXA8VJiS5ap43JXiUFFAaQ=="
        )
        options = 'command="/usr/local/lib/ssh-helper"'
        email = "github.com"
        empty_line = "\n"
        comment_line = "# this is a comment\n"

        # Write out the authorized key to a temporary file
        temp_file = tempfile.NamedTemporaryFile(delete=False, mode="w+")
        temp_file.close()

        with salt.utils.files.fopen(temp_file.name, "w") as _fh:
            # Add comment
            _fh.write(comment_line)
            # Add empty line for #41335
            _fh.write(empty_line)
            _fh.write(f"{options} {enc} {key} {email}")

        with patch.dict(ssh.__salt__, {"user.info": MagicMock(return_value={})}):
            with patch(
                "salt.modules.ssh._get_config_file",
                MagicMock(return_value=temp_file.name),
            ):
                ssh._replace_auth_key("foo", key, config=temp_file.name)

        # The previous authorized key should have been replaced by the simpler one
        with salt.utils.files.fopen(temp_file.name) as _fh:
            file_txt = salt.utils.stringutils.to_unicode(_fh.read())
            self.assertIn(enc, file_txt)
            self.assertIn(key, file_txt)
            self.assertNotIn(options, file_txt)
            self.assertNotIn(email, file_txt)

        # Now test a very simple key using ecdsa instead of ssh-rsa and with multiple options
        enc = "ecdsa-sha2-nistp256"
        key = "abcxyz"

        with salt.utils.files.fopen(temp_file.name, "a") as _fh:
            _fh.write(salt.utils.stringutils.to_str(f"{enc} {key}"))

        # Replace the simple key from before with the more complicated options + new email
        # Option example is taken from Pull Request #39855
        options = [
            "no-port-forwarding",
            "no-agent-forwarding",
            "no-X11-forwarding",
            'command="echo \'Please login as the user "ubuntu" rather than the user'
            ' "root".\'',
        ]
        email = "foo@example.com"

        with patch.dict(ssh.__salt__, {"user.info": MagicMock(return_value={})}):
            with patch(
                "salt.modules.ssh._get_config_file",
                MagicMock(return_value=temp_file.name),
            ):
                ssh._replace_auth_key(
                    "foo",
                    key,
                    enc=enc,
                    comment=email,
                    options=options,
                    config=temp_file.name,
                )

        # Assert that the new line was added as-is to the file
        with salt.utils.files.fopen(temp_file.name) as _fh:
            file_txt = salt.utils.stringutils.to_unicode(_fh.read())
            self.assertIn(enc, file_txt)
            self.assertIn(key, file_txt)
            self.assertIn("{} ".format(",".join(options)), file_txt)
            self.assertIn(email, file_txt)
            self.assertIn(empty_line, file_txt)
            self.assertIn(comment_line, file_txt)

        # Now test a another very simple key using sk-ed25519 instead of ssh-rsa and with
        # multiple options
        enc = "sk-ssh-ed25519-cert-v01@openssh.com"
        key = "abcxyz"

        with salt.utils.files.fopen(temp_file.name, "a") as _fh:
            _fh.write(salt.utils.stringutils.to_str(f"{enc} {key}"))

        # Replace the simple key from before with the more complicated options + new email
        # Option example is taken from Pull Request #39855
        options = [
            "no-agent-forwarding",
            'command="echo \'Please login as the user "debian" rather than the user'
            ' "root".\'',
        ]
        email = "foobazbar@example.com"

        with patch.dict(ssh.__salt__, {"user.info": MagicMock(return_value={})}):
            with patch(
                "salt.modules.ssh._get_config_file",
                MagicMock(return_value=temp_file.name),
            ):
                ssh._replace_auth_key(
                    "foo",
                    key,
                    enc=enc,
                    comment=email,
                    options=options,
                    config=temp_file.name,
                )

        with salt.utils.files.fopen(temp_file.name) as _fh:
            file_txt = salt.utils.stringutils.to_unicode(_fh.read())
            # the initial key must have been replaced and no longer present
            self.assertNotIn(f"\n{enc} {key}\n", file_txt)
            # the new key must be present
            self.assertIn(
                "{} {} {} {}".format(",".join(options), enc, key, email), file_txt
            )
            self.assertIn(empty_line, file_txt)
            self.assertIn(comment_line, file_txt)

    def test_rm_auth_key(self):
        """
        Test the rm_auth_key with some different authorized_keys examples
        """

        temp_file = tempfile.NamedTemporaryFile(delete=False, mode="w+")
        temp_file.close()

        options = 'command="echo command"'
        enc = "sk-ssh-ed25519-cert-v01@openssh.com"
        key = "abcxyz"
        email = "foo@example.com"

        with salt.utils.files.fopen(temp_file.name, "w") as _fh:
            _fh.write(salt.utils.stringutils.to_str(f"{options} {enc} {key} {email}"))

        with patch.dict(ssh.__salt__, {"user.info": MagicMock(return_value={})}):
            with patch(
                "salt.modules.ssh._get_config_file",
                MagicMock(return_value=temp_file.name),
            ):
                with patch(
                    "salt.modules.ssh.auth_keys",
                    MagicMock(
                        return_value=salt.utils.files.fopen(temp_file.name, "r").read()
                    ),
                ):
                    self.assertEqual(
                        ssh.rm_auth_key(
                            "foo",
                            f"{key}",
                            config=temp_file.name,
                        ),
                        "Key removed",
                    )

        # Assert that the key was removed
        with salt.utils.files.fopen(temp_file.name) as _fh:
            file_txt = salt.utils.stringutils.to_unicode(_fh.read())
            self.assertNotIn(key, file_txt)



"""
    Test cases for salt.pillar.etcd_pillar

    :codeauthor: Caleb Beard <calebb@vmware.com>
"""

import pytest

import salt.pillar.etcd_pillar as etcd_pillar
import salt.utils.etcd_util as etcd_util
from tests.support.mock import MagicMock, create_autospec, patch


@pytest.fixture
def configure_loader_modules():
    return {etcd_pillar: {}}


@pytest.fixture
def instance():
    return create_autospec(etcd_util.EtcdBase)


@pytest.fixture
def etcd_client_mock(instance):
    mocked_client = MagicMock()
    mocked_client.return_value = instance
    return mocked_client


def test_ext_pillar(etcd_client_mock, instance):
    """
    Test ext_pillar functionality
    """
    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock):
        # Test pillar with no root given
        instance.tree.return_value = {"key": "value"}
        assert etcd_pillar.ext_pillar("test-id", {}, "etcd_profile") == {"key": "value"}
        instance.tree.assert_called_with("/")

        # Test pillar with a root given
        instance.tree.return_value = {"key": "value"}
        assert etcd_pillar.ext_pillar("test-id", {}, "etcd_profile root=/salt") == {
            "key": "value"
        }
        instance.tree.assert_called_with("/salt")

        # Test pillar with a root given that uses the minion id
        instance.tree.return_value = {"key": "value"}
        assert etcd_pillar.ext_pillar(
            "test-id", {}, "etcd_profile root=/salt/%(minion_id)s"
        ) == {"key": "value"}
        instance.tree.assert_called_with("/salt/test-id")

        # Test pillar with a root given that uses the minion id
        instance.tree.side_effect = KeyError
        assert etcd_pillar.ext_pillar("test-id", {"key": "value"}, "etcd_profile") == {}



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.smf
"""

import pytest

import salt.modules.smf_service as smf
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {smf: {}}


def test_get_running():
    """
    Test to return the running services
    """
    with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A online\n")}):
        assert smf.get_running() == ["A"]


def test_get_stopped():
    """
    Test to return the stopped services
    """
    with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A\n")}):
        assert smf.get_stopped() == ["A"]


def test_available():
    """
    Test to returns ``True`` if the specified service is available,
    otherwise returns ``False``.
    """
    with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A")}):
        with patch.object(smf, "get_all", return_value="A"):
            assert smf.available("A")


def test_missing():
    """
    The inverse of service.available.
    Returns ``True`` if the specified service is not available, otherwise
    returns ``False``.
    """
    with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A")}):
        with patch.object(smf, "get_all", return_value="A"):
            assert not smf.missing("A")


def test_get_all():
    """
    Test to return all installed services
    """
    with patch.dict(smf.__salt__, {"cmd.run": MagicMock(return_value="A\n")}):
        assert smf.get_all() == ["A"]


def test_start():
    """
    Test to start the specified service
    """
    with patch.dict(
        smf.__salt__,
        {"cmd.retcode": MagicMock(side_effect=[False, 3, None, False, 4])},
    ):
        assert smf.start("name")

        assert smf.start("name")

        assert not smf.start("name")


def test_stop():
    """
    Test to stop the specified service
    """
    with patch.dict(smf.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
        assert smf.stop("name")


def test_restart():
    """
    Test to restart the named service
    """
    with patch.dict(
        smf.__salt__, {"cmd.retcode": MagicMock(side_effect=[False, True])}
    ):

        with patch.object(smf, "start", return_value="A"):
            assert smf.restart("name") == "A"

        assert not smf.restart("name")


def test_reload_():
    """
    Test to reload the named service
    """
    with patch.dict(
        smf.__salt__, {"cmd.retcode": MagicMock(side_effect=[False, True])}
    ):

        with patch.object(smf, "start", return_value="A"):
            assert smf.reload_("name") == "A"

        assert not smf.reload_("name")


def test_status():
    """
    Test to return the status for a service, returns a bool whether the
    service is running.
    """
    with patch.dict(
        smf.__salt__, {"cmd.run": MagicMock(side_effect=["online", "online1"])}
    ):
        assert smf.status("name")

        assert not smf.status("name")


def test_enable():
    """
    Test to enable the named service to start at boot
    """
    with patch.dict(smf.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
        assert smf.enable("name")


def test_disable():
    """
    Test to disable the named service to start at boot
    """
    with patch.dict(smf.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
        assert smf.disable("name")


def test_enabled():
    """
    Test to check to see if the named service is enabled to start on boot
    """
    with patch.dict(
        smf.__salt__,
        {"cmd.run": MagicMock(side_effect=["fmri", "A B true", "fmri", "A B false"])},
    ):
        assert smf.enabled("name")

        assert not smf.enabled("name")


def test_disabled():
    """
    Test to check to see if the named service is disabled to start on boot
    """
    with patch.object(smf, "enabled", return_value=False):
        assert smf.disabled("name")


def test_get_enabled():
    """
    Test to return the enabled services
    """
    with patch.object(smf, "_get_enabled_disabled", return_value=True):
        assert smf.get_enabled()


def test_get_disabled():
    """
    Test to return the disabled services
    """
    with patch.object(smf, "_get_enabled_disabled", return_value=True):
        assert smf.get_disabled()



"""
Runner for SmartOS minions control vmadm
"""

from collections import OrderedDict

import salt.client
from salt.exceptions import SaltClientError

# Function aliases
__func_alias__ = {"list_vms": "list"}

# Define the module's virtual name
__virtualname__ = "vmadm"


def __virtual__():
    """
    Provides vmadm runner
    """
    # NOTE: always load vmadm runner
    #       we could check using test.ping + a grain
    #       match, but doing this on master startup is
    #       not acceptable
    return __virtualname__


def _action(action="get", search=None, one=True, force=False):
    """
    Multi action helper for start, stop, get, ...
    """
    vms = {}
    matched_vms = []

    ## lookup vms
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        try:
            vmadm_args = {}
            vmadm_args["order"] = "uuid,alias,hostname,state"
            if "=" in search:
                vmadm_args["search"] = search
            for cn in client.cmd_iter(
                "G@virtual:physical and G@os:smartos",
                "vmadm.list",
                kwarg=vmadm_args,
                tgt_type="compound",
            ):
                if not cn:
                    continue
                node = next(iter(cn.keys()))
                if (
                    not isinstance(cn[node], dict)
                    or "ret" not in cn[node]
                    or not isinstance(cn[node]["ret"], dict)
                ):
                    continue
                for vm in cn[node]["ret"]:
                    vmcfg = cn[node]["ret"][vm]
                    vmcfg["node"] = node
                    vms[vm] = vmcfg
        except SaltClientError as client_error:
            pass

        ## check if we have vms
        if len(vms) == 0:
            return {"Error": "No vms found."}

        ## simple search
        if "=" not in search:
            loop_pass = 0
            while loop_pass < 3:
                ## each pass will try a different field
                if loop_pass == 0:
                    field = "uuid"
                elif loop_pass == 1:
                    field = "hostname"
                else:
                    field = "alias"

                ## loop vms and try to match
                for vm in vms:
                    if field == "uuid" and vm == search:
                        matched_vms.append(vm)
                        break  # exit for on uuid match (max = 1)
                    elif field in vms[vm] and vms[vm][field] == search:
                        matched_vms.append(vm)

                ## exit on match(es) or try again
                if len(matched_vms) > 0:
                    break
                else:
                    loop_pass += 1
        else:
            for vm in vms:
                matched_vms.append(vm)

        ## check if we have vms
        if len(matched_vms) == 0:
            return {"Error": "No vms matched."}

        ## multiple allowed?
        if one and len(matched_vms) > 1:
            return {
                "Error": f"Matched {len(matched_vms)} vms, only one allowed!",
                "Matches": matched_vms,
            }

        ## perform action
        ret = {}
        if action in ["start", "stop", "reboot", "get"]:
            for vm in matched_vms:
                vmadm_args = {"key": "uuid", "vm": vm}
                try:
                    for vmadm_res in client.cmd_iter(
                        vms[vm]["node"], f"vmadm.{action}", kwarg=vmadm_args
                    ):
                        if not vmadm_res:
                            continue
                        if vms[vm]["node"] in vmadm_res:
                            ret[vm] = vmadm_res[vms[vm]["node"]]["ret"]
                except SaltClientError as client_error:
                    ret[vm] = False
        elif action in ["is_running"]:
            ret = True
            for vm in matched_vms:
                if vms[vm]["state"] != "running":
                    ret = False
                    break
        return ret


def nodes(verbose=False):
    """
    List all compute nodes

    verbose : boolean
        print additional information about the node
        e.g. platform version, hvm capable, ...

    CLI Example:

    .. code-block:: bash

        salt-run vmadm.nodes
        salt-run vmadm.nodes verbose=True
    """
    ret = {} if verbose else []
    with salt.client.get_local_client(__opts__["conf_file"]) as client:

        ## get list of nodes
        try:
            for cn in client.cmd_iter(
                "G@virtual:physical and G@os:smartos",
                "grains.items",
                tgt_type="compound",
            ):
                if not cn:
                    continue
                node = next(iter(cn.keys()))
                if (
                    not isinstance(cn[node], dict)
                    or "ret" not in cn[node]
                    or not isinstance(cn[node]["ret"], dict)
                ):
                    continue
                if verbose:
                    ret[node] = {}
                    ret[node]["version"] = {}
                    ret[node]["version"]["platform"] = cn[node]["ret"]["osrelease"]
                    if "computenode_sdc_version" in cn[node]["ret"]:
                        ret[node]["version"]["sdc"] = cn[node]["ret"][
                            "computenode_sdc_version"
                        ]
                    ret[node]["vms"] = {}
                    if (
                        "computenode_vm_capable" in cn[node]["ret"]
                        and cn[node]["ret"]["computenode_vm_capable"]
                        and "computenode_vm_hw_virt" in cn[node]["ret"]
                    ):
                        ret[node]["vms"]["hw_cap"] = cn[node]["ret"][
                            "computenode_vm_hw_virt"
                        ]
                    else:
                        ret[node]["vms"]["hw_cap"] = False
                    if "computenode_vms_running" in cn[node]["ret"]:
                        ret[node]["vms"]["running"] = cn[node]["ret"][
                            "computenode_vms_running"
                        ]
                else:
                    ret.append(node)
        except SaltClientError as client_error:
            return f"{client_error}"

        if not verbose:
            ret.sort()
        return ret


def list_vms(search=None, verbose=False):
    """
    List all vms

    search : string
        filter vms, see the execution module
    verbose : boolean
        print additional information about the vm

    CLI Example:

    .. code-block:: bash

        salt-run vmadm.list
        salt-run vmadm.list search='type=KVM'
        salt-run vmadm.list verbose=True
    """
    ret = OrderedDict() if verbose else []
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        try:
            vmadm_args = {}
            vmadm_args["order"] = "uuid,alias,hostname,state,type,cpu_cap,vcpus,ram"
            if search:
                vmadm_args["search"] = search
            for cn in client.cmd_iter(
                "G@virtual:physical and G@os:smartos",
                "vmadm.list",
                kwarg=vmadm_args,
                tgt_type="compound",
            ):
                if not cn:
                    continue
                node = next(iter(cn.keys()))
                if (
                    not isinstance(cn[node], dict)
                    or "ret" not in cn[node]
                    or not isinstance(cn[node]["ret"], dict)
                ):
                    continue
                for vm in cn[node]["ret"]:
                    vmcfg = cn[node]["ret"][vm]
                    if verbose:
                        ret[vm] = OrderedDict()
                        ret[vm]["hostname"] = vmcfg["hostname"]
                        ret[vm]["alias"] = vmcfg["alias"]
                        ret[vm]["computenode"] = node
                        ret[vm]["state"] = vmcfg["state"]
                        ret[vm]["resources"] = OrderedDict()
                        ret[vm]["resources"]["memory"] = vmcfg["ram"]
                        if vmcfg["type"] == "KVM":
                            ret[vm]["resources"]["cpu"] = "{:.2f}".format(
                                int(vmcfg["vcpus"])
                            )
                        else:
                            if vmcfg["cpu_cap"] != "":
                                ret[vm]["resources"]["cpu"] = "{:.2f}".format(
                                    int(vmcfg["cpu_cap"]) / 100
                                )
                    else:
                        ret.append(vm)
        except SaltClientError as client_error:
            return f"{client_error}"

        if not verbose:
            ret = sorted(ret)

        return ret


def start(search, one=True):
    """
    Start one or more vms

    search : string
        filter vms, see the execution module.
    one : boolean
        start only one vm

    .. note::
        If the search parameter does not contain an equal (=) symbol it will be
        assumed it will be tried as uuid, hostname, and alias.

    CLI Example:

    .. code-block:: bash

        salt-run vmadm.start 91244bba-1146-e4ec-c07e-e825e0223aa9
        salt-run vmadm.start search='alias=jiska'
        salt-run vmadm.start search='type=KVM' one=False
    """
    return _action("start", search, one)


def stop(search, one=True):
    """
    Stop one or more vms

    search : string
        filter vms, see the execution module.
    one : boolean
        stop only one vm

    .. note::
        If the search parameter does not contain an equal (=) symbol it will be
        assumed it will be tried as uuid, hostname, and alias.

    CLI Example:

    .. code-block:: bash

        salt-run vmadm.stop 91244bba-1146-e4ec-c07e-e825e0223aa9
        salt-run vmadm.stop search='alias=jody'
        salt-run vmadm.stop search='type=KVM' one=False
    """
    return _action("stop", search, one)


def reboot(search, one=True, force=False):
    """
    Reboot one or more vms

    search : string
        filter vms, see the execution module.
    one : boolean
        reboot only one vm
    force : boolean
        force reboot, faster but no graceful shutdown

    .. note::
        If the search parameter does not contain an equal (=) symbol it will be
        assumed it will be tried as uuid, hostname, and alias.

    CLI Example:

    .. code-block:: bash

        salt-run vmadm.reboot 91244bba-1146-e4ec-c07e-e825e0223aa9
        salt-run vmadm.reboot search='alias=marije'
        salt-run vmadm.reboot search='type=KVM' one=False
    """
    return _action("reboot", search, one, force)


def get(search, one=True):
    """
    Return information for vms

    search : string
        filter vms, see the execution module.
    one : boolean
        return only one vm

    .. note::
        If the search parameter does not contain an equal (=) symbol it will be
        assumed it will be tried as uuid, hostname, and alias.

    CLI Example:

    .. code-block:: bash

        salt-run vmadm.get 91244bba-1146-e4ec-c07e-e825e0223aa9
        salt-run vmadm.get search='alias=saskia'
    """
    return _action("get", search, one)


def is_running(search):
    """
    Return true if vm is running

    search : string
        filter vms, see the execution module.

    .. note::
        If the search parameter does not contain an equal (=) symbol it will be
        assumed it will be tried as uuid, hostname, and alias.

    .. note::
        If multiple vms are matched, the result will be true of ALL vms are running

    CLI Example:

    .. code-block:: bash

        salt-run vmadm.is_running 91244bba-1146-e4ec-c07e-e825e0223aa9
        salt-run vmadm.is_running search='alias=julia'
    """
    return _action("is_running", search, False)



#!/usr/bin/env python
#
# Copyright 2012 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# pylint: skip-file
from __future__ import absolute_import, division, print_function

import contextlib
import glob
import logging
import os
import re
import subprocess
import sys
import tempfile
import warnings

from salt.ext.tornado.escape import utf8
from salt.ext.tornado.log import LogFormatter, define_logging_options, enable_pretty_logging
from salt.ext.tornado.options import OptionParser
from salt.ext.tornado.test.util import unittest
from salt.ext.tornado.util import basestring_type


@contextlib.contextmanager
def ignore_bytes_warning():
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=BytesWarning)
        yield


class LogFormatterTest(unittest.TestCase):
    # Matches the output of a single logging call (which may be multiple lines
    # if a traceback was included, so we use the DOTALL option)
    LINE_RE = re.compile(b"(?s)\x01\\[E [0-9]{6} [0-9]{2}:[0-9]{2}:[0-9]{2} log_test:[0-9]+\\]\x02 (.*)")

    def setUp(self):
        self.formatter = LogFormatter(color=False)
        # Fake color support.  We can't guarantee anything about the $TERM
        # variable when the tests are run, so just patch in some values
        # for testing.  (testing with color off fails to expose some potential
        # encoding issues from the control characters)
        self.formatter._colors = {
            logging.ERROR: u"\u0001",
        }
        self.formatter._normal = u"\u0002"
        # construct a Logger directly to bypass getLogger's caching
        self.logger = logging.Logger('LogFormatterTest')
        self.logger.propagate = False
        self.tempdir = tempfile.mkdtemp()
        self.filename = os.path.join(self.tempdir, 'log.out')
        self.handler = self.make_handler(self.filename)
        self.handler.setFormatter(self.formatter)
        self.logger.addHandler(self.handler)

    def tearDown(self):
        self.handler.close()
        os.unlink(self.filename)
        os.rmdir(self.tempdir)

    def make_handler(self, filename):
        # Base case: default setup without explicit encoding.
        # In python 2, supports arbitrary byte strings and unicode objects
        # that contain only ascii.  In python 3, supports ascii-only unicode
        # strings (but byte strings will be repr'd automatically).
        return logging.FileHandler(filename)

    def get_output(self):
        with open(self.filename, "rb") as f:
            line = f.read().strip()
            m = LogFormatterTest.LINE_RE.match(line)
            if m:
                return m.group(1)
            else:
                raise Exception("output didn't match regex: %r" % line)

    def test_basic_logging(self):
        self.logger.error("foo")
        self.assertEqual(self.get_output(), b"foo")

    def test_bytes_logging(self):
        with ignore_bytes_warning():
            # This will be "\xe9" on python 2 or "b'\xe9'" on python 3
            self.logger.error(b"\xe9")
            self.assertEqual(self.get_output(), utf8(repr(b"\xe9")))

    def test_utf8_logging(self):
        with ignore_bytes_warning():
            self.logger.error(u"\u00e9".encode("utf8"))
        if issubclass(bytes, basestring_type):
            # on python 2, utf8 byte strings (and by extension ascii byte
            # strings) are passed through as-is.
            self.assertEqual(self.get_output(), utf8(u"\u00e9"))
        else:
            # on python 3, byte strings always get repr'd even if
            # they're ascii-only, so this degenerates into another
            # copy of test_bytes_logging.
            self.assertEqual(self.get_output(), utf8(repr(utf8(u"\u00e9"))))

    def test_bytes_exception_logging(self):
        try:
            raise Exception(b'\xe9')
        except Exception:
            self.logger.exception('caught exception')
        # This will be "Exception: \xe9" on python 2 or
        # "Exception: b'\xe9'" on python 3.
        output = self.get_output()
        self.assertRegexpMatches(output, br'Exception.*\\xe9')
        # The traceback contains newlines, which should not have been escaped.
        self.assertNotIn(br'\n', output)


class UnicodeLogFormatterTest(LogFormatterTest):
    def make_handler(self, filename):
        # Adding an explicit encoding configuration allows non-ascii unicode
        # strings in both python 2 and 3, without changing the behavior
        # for byte strings.
        return logging.FileHandler(filename, encoding="utf8")

    def test_unicode_logging(self):
        self.logger.error(u"\u00e9")
        self.assertEqual(self.get_output(), utf8(u"\u00e9"))


class EnablePrettyLoggingTest(unittest.TestCase):
    def setUp(self):
        super(EnablePrettyLoggingTest, self).setUp()
        self.options = OptionParser()
        define_logging_options(self.options)
        self.logger = logging.Logger('tornado.test.log_test.EnablePrettyLoggingTest')
        self.logger.propagate = False

    def test_log_file(self):
        tmpdir = tempfile.mkdtemp()
        try:
            self.options.log_file_prefix = tmpdir + '/test_log'
            enable_pretty_logging(options=self.options, logger=self.logger)
            self.assertEqual(1, len(self.logger.handlers))
            self.logger.error('hello')
            self.logger.handlers[0].flush()
            filenames = glob.glob(tmpdir + '/test_log*')
            self.assertEqual(1, len(filenames))
            with open(filenames[0]) as f:
                self.assertRegexpMatches(f.read(), r'^\[E [^]]*\] hello$')
        finally:
            for handler in self.logger.handlers:
                handler.flush()
                handler.close()
            for filename in glob.glob(tmpdir + '/test_log*'):
                os.unlink(filename)
            os.rmdir(tmpdir)

    def test_log_file_with_timed_rotating(self):
        tmpdir = tempfile.mkdtemp()
        try:
            self.options.log_file_prefix = tmpdir + '/test_log'
            self.options.log_rotate_mode = 'time'
            enable_pretty_logging(options=self.options, logger=self.logger)
            self.logger.error('hello')
            self.logger.handlers[0].flush()
            filenames = glob.glob(tmpdir + '/test_log*')
            self.assertEqual(1, len(filenames))
            with open(filenames[0]) as f:
                self.assertRegexpMatches(
                    f.read(),
                    r'^\[E [^]]*\] hello$')
        finally:
            for handler in self.logger.handlers:
                handler.flush()
                handler.close()
            for filename in glob.glob(tmpdir + '/test_log*'):
                os.unlink(filename)
            os.rmdir(tmpdir)

    def test_wrong_rotate_mode_value(self):
        try:
            self.options.log_file_prefix = 'some_path'
            self.options.log_rotate_mode = 'wrong_mode'
            self.assertRaises(ValueError, enable_pretty_logging,
                              options=self.options, logger=self.logger)
        finally:
            for handler in self.logger.handlers:
                handler.flush()
                handler.close()


class LoggingOptionTest(unittest.TestCase):
    """Test the ability to enable and disable Tornado's logging hooks."""
    def logs_present(self, statement, args=None):
        # Each test may manipulate and/or parse the options and then logs
        # a line at the 'info' level.  This level is ignored in the
        # logging module by default, but Tornado turns it on by default
        # so it is the easiest way to tell whether tornado's logging hooks
        # ran.
        IMPORT = 'from salt.ext.tornado.options import options, parse_command_line'
        LOG_INFO = 'import logging; logging.info("hello")'
        program = ';'.join([IMPORT, statement, LOG_INFO])
        proc = subprocess.Popen(
            [sys.executable, '-c', program] + (args or []),
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0, 'process failed: %r' % stdout)
        return b'hello' in stdout

    def test_default(self):
        self.assertFalse(self.logs_present('pass'))

    def test_tornado_default(self):
        self.assertTrue(self.logs_present('parse_command_line()'))

    def test_disable_command_line(self):
        self.assertFalse(self.logs_present('parse_command_line()',
                                           ['--logging=none']))

    def test_disable_command_line_case_insensitive(self):
        self.assertFalse(self.logs_present('parse_command_line()',
                                           ['--logging=None']))

    def test_disable_code_string(self):
        self.assertFalse(self.logs_present(
            'options.logging = "none"; parse_command_line()'))

    def test_disable_code_none(self):
        self.assertFalse(self.logs_present(
            'options.logging = None; parse_command_line()'))

    def test_disable_override(self):
        # command line trumps code defaults
        self.assertTrue(self.logs_present(
            'options.logging = None; parse_command_line()',
            ['--logging=info']))



"""
Management of OpenStack Neutron Subnets
=========================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.neutronng` for setup instructions

Example States

.. code-block:: yaml

    create subnet:
      neutron_subnet.present:
        - name: subnet1
        - network_name_or_id: network1
        - cidr: 192.168.199.0/24


    delete subnet:
      neutron_subnet.absent:
        - name: subnet2

    create subnet with optional params:
      neutron_subnet.present:
        - name: subnet1
        - network_name_or_id: network1
        - enable_dhcp: True
        - cidr: 192.168.199.0/24
        - allocation_pools:
          - start: 192.168.199.5
            end: 192.168.199.250
        - host_routes:
          - destination: 192.168..0.0/24
            nexthop: 192.168.0.1
        - gateway_ip: 192.168.199.1
        - dns_nameservers:
          - 8.8.8.8
          - 8.8.8.7

    create ipv6 subnet:
      neutron_subnet.present:
        - name: v6subnet1
        - network_name_or_id: network1
        - ip_version: 6
"""

__virtualname__ = "neutron_subnet"


def __virtual__():
    if "neutronng.list_subnets" in __salt__:
        return __virtualname__
    return (
        False,
        "The neutronng execution module failed to load: shade python module is not available",
    )


def present(name, auth=None, **kwargs):
    """
    Ensure a subnet exists and is up-to-date

    name
        Name of the subnet

    network_name_or_id
        The unique name or ID of the attached network.
        If a non-unique name is supplied, an exception is raised.

    allocation_pools
        A list of dictionaries of the start and end addresses
        for the allocation pools

    gateway_ip
        The gateway IP address.

    dns_nameservers
        A list of DNS name servers for the subnet.

    host_routes
        A list of host route dictionaries for the subnet.

    ipv6_ra_mode
        IPv6 Router Advertisement mode.
        Valid values are: ‘dhcpv6-stateful’, ‘dhcpv6-stateless’, or ‘slaac’.

    ipv6_address_mode
        IPv6 address mode.
        Valid values are: ‘dhcpv6-stateful’, ‘dhcpv6-stateless’, or ‘slaac’.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["neutronng.setup_clouds"](auth)

    kwargs["subnet_name"] = name
    subnet = __salt__["neutronng.subnet_get"](name=name)

    if subnet is None:
        if __opts__["test"]:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["comment"] = "Subnet will be created."
            return ret

        new_subnet = __salt__["neutronng.subnet_create"](**kwargs)
        ret["changes"] = new_subnet
        ret["comment"] = "Created subnet"
        return ret

    changes = __salt__["neutronng.compare_changes"](subnet, **kwargs)
    if changes:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = changes
            ret["comment"] = "Project will be updated."
            return ret

        # update_subnet does not support changing cidr,
        # so we have to delete and recreate the subnet in this case.
        if "cidr" in changes or "tenant_id" in changes:
            __salt__["neutronng.subnet_delete"](name=name)
            new_subnet = __salt__["neutronng.subnet_create"](**kwargs)
            ret["changes"] = new_subnet
            ret["comment"] = "Deleted and recreated subnet"
            return ret

        __salt__["neutronng.subnet_update"](**kwargs)
        ret["changes"].update(changes)
        ret["comment"] = "Updated subnet"

    return ret


def absent(name, auth=None):
    """
    Ensure a subnet does not exists

    name
        Name of the subnet

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    __salt__["neutronng.setup_clouds"](auth)

    subnet = __salt__["neutronng.subnet_get"](name=name)

    if subnet:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = {"id": subnet.id}
            ret["comment"] = "Project will be deleted."
            return ret

        __salt__["neutronng.subnet_delete"](name=subnet)
        ret["changes"]["id"] = name
        ret["comment"] = "Deleted subnet"

    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import socket

import pytest

import salt.modules.win_network as win_network
import salt.utils.network
from tests.support.mock import MagicMock, Mock, patch

try:
    import wmi

    HAS_WMI = True
except ImportError:
    HAS_WMI = False


class Mockwmi:
    """
    Mock wmi class
    """

    NetConnectionID = "Ethernet"

    def __init__(self):
        pass


class Mockwinapi:
    """
    Mock winapi class
    """

    def __init__(self):
        pass

    class winapi:
        """
        Mock winapi class
        """

        def __init__(self):
            pass

        class Com:
            """
            Mock Com method
            """

            def __enter__(self):
                return self

            def __exit__(self, *exc_info):
                return False


@pytest.fixture
def configure_loader_modules():
    return {win_network: {}}


def test_ping():
    """
    Test if it performs a ping to a host.
    """
    mock = MagicMock(return_value=True)
    with patch.dict(win_network.__salt__, {"cmd.run": mock}):
        assert win_network.ping("127.0.0.1")


def test_netstat():
    """
    Test if it return information on open ports and states
    """
    ret = (
        "  Proto  Local Address    Foreign Address    State    PID\n"
        "  TCP    127.0.0.1:1434    0.0.0.0:0    LISTENING    1728\n"
        "  UDP    127.0.0.1:1900    *:*        4240"
    )
    mock = MagicMock(return_value=ret)
    with patch.dict(win_network.__salt__, {"cmd.run": mock}):
        assert win_network.netstat() == [
            {
                "local-address": "127.0.0.1:1434",
                "program": "1728",
                "proto": "TCP",
                "remote-address": "0.0.0.0:0",
                "state": "LISTENING",
            },
            {
                "local-address": "127.0.0.1:1900",
                "program": "4240",
                "proto": "UDP",
                "remote-address": "*:*",
                "state": None,
            },
        ]


def test_traceroute():
    """
    Test if it performs a traceroute to a 3rd party host
    """
    ret = (
        "  1     1 ms    <1 ms    <1 ms  172.27.104.1\n"
        "  2     1 ms    <1 ms     1 ms  121.242.35.1.s[121.242.35.1]\n"
        "  3     3 ms     2 ms     2 ms  121.242.4.53.s[121.242.4.53]\n"
    )
    mock = MagicMock(return_value=ret)
    with patch.dict(win_network.__salt__, {"cmd.run": mock}):
        assert win_network.traceroute("google.com") == [
            {
                "count": "1",
                "hostname": None,
                "ip": "172.27.104.1",
                "ms1": "1",
                "ms2": "<1",
                "ms3": "<1",
            },
            {
                "count": "2",
                "hostname": None,
                "ip": "121.242.35.1.s[121.242.35.1]",
                "ms1": "1",
                "ms2": "<1",
                "ms3": "1",
            },
            {
                "count": "3",
                "hostname": None,
                "ip": "121.242.4.53.s[121.242.4.53]",
                "ms1": "3",
                "ms2": "2",
                "ms3": "2",
            },
        ]


def test_nslookup():
    """
    Test if it query DNS for information about a domain or ip address
    """
    ret = (
        "Server:  ct-dc-3-2.cybage.com\n"
        "Address:  172.27.172.12\n"
        "Non-authoritative answer:\n"
        "Name:    google.com\n"
        "Addresses:  2404:6800:4007:806::200e\n"
        "216.58.196.110\n"
    )
    mock = MagicMock(return_value=ret)
    with patch.dict(win_network.__salt__, {"cmd.run": mock}):
        assert win_network.nslookup("google.com") == [
            {"Server": "ct-dc-3-2.cybage.com"},
            {"Address": "172.27.172.12"},
            {"Name": "google.com"},
            {"Addresses": ["2404:6800:4007:806::200e", "216.58.196.110"]},
        ]


def test_dig():
    """
    Test if it performs a DNS lookup with dig
    """
    mock = MagicMock(return_value=True)
    with patch.dict(win_network.__salt__, {"cmd.run": mock}):
        assert win_network.dig("google.com")


@pytest.mark.skipif(HAS_WMI is False, reason="WMI is available only on Windows")
def test_interfaces_names():
    """
    Test if it return a list of all the interfaces names
    """
    WMI = Mock()
    WMI.Win32_NetworkAdapter = MagicMock(return_value=Mockwmi)
    with patch("salt.utils.winapi.Com", MagicMock()), patch.object(
        WMI, "Win32_NetworkAdapter", return_value=[Mockwmi()]
    ), patch("salt.utils", Mockwinapi), patch.object(
        wmi, "WMI", Mock(return_value=WMI)
    ):
        assert win_network.interfaces_names() == ["Ethernet"]


def test_interfaces():
    """
    Test if it return information about all the interfaces on the minion
    """
    with patch.object(
        salt.utils.network, "win_interfaces", MagicMock(return_value=True)
    ):
        assert win_network.interfaces()


def test_hw_addr():
    """
    Test if it return the hardware address (a.k.a. MAC address)
    for a given interface
    """
    with patch.object(
        salt.utils.network, "hw_addr", MagicMock(return_value="Ethernet")
    ):
        assert win_network.hw_addr("Ethernet") == "Ethernet"


def test_subnets():
    """
    Test if it returns a list of subnets to which the host belongs
    """
    with patch.object(
        salt.utils.network, "subnets", MagicMock(return_value="10.1.1.0/24")
    ):
        assert win_network.subnets() == "10.1.1.0/24"


def test_in_subnet():
    """
    Test if it returns True if host is within specified subnet,
    otherwise False
    """
    with patch.object(salt.utils.network, "in_subnet", MagicMock(return_value=True)):
        assert win_network.in_subnet("10.1.1.0/16")


def test_get_route():
    """
    Test if it return information on open ports and states
    """
    ret = (
        "\n\n"
        "IPAddress         : 10.0.0.15\n"
        "InterfaceIndex    : 3\n"
        "InterfaceAlias    : Wi-Fi\n"
        "AddressFamily     : IPv4\n"
        "Type              : Unicast\n"
        "PrefixLength      : 24\n"
        "PrefixOrigin      : Dhcp\n"
        "SuffixOrigin      : Dhcp\n"
        "AddressState      : Preferred\n"
        "ValidLifetime     : 6.17:52:39\n"
        "PreferredLifetime : 6.17:52:39\n"
        "SkipAsSource      : False\n"
        "PolicyStore       : ActiveStore\n"
        "\n\n"
        "Caption            :\n"
        "Description        :\n"
        "ElementName        :\n"
        "InstanceID         : :8:8:8:9:55=55;:8;8;:8;55;\n"
        "AdminDistance      :\n"
        "DestinationAddress :\n"
        "IsStatic           :\n"
        "RouteMetric        : 0\n"
        "TypeOfRoute        : 3\n"
        "AddressFamily      : IPv4\n"
        "CompartmentId      : 1\n"
        "DestinationPrefix  : 0.0.0.0/0\n"
        "InterfaceAlias     : Wi-Fi\n"
        "InterfaceIndex     : 3\n"
        "NextHop            : 10.0.0.1\n"
        "PreferredLifetime  : 6.23:14:43\n"
        "Protocol           : NetMgmt\n"
        "Publish            : No\n"
        "Store              : ActiveStore\n"
        "ValidLifetime      : 6.23:14:43\n"
        "PSComputerName     :\n"
        "ifIndex            : 3"
    )
    mock = MagicMock(return_value=ret)
    with patch.dict(win_network.__salt__, {"cmd.run": mock}):
        assert win_network.get_route("192.0.0.8") == {
            "destination": "192.0.0.8",
            "gateway": "10.0.0.1",
            "interface": "Wi-Fi",
            "source": "10.0.0.15",
        }


def test_connect_53371():
    """
    Test that UnboundLocalError is not thrown on socket.gaierror
    as reported in #53371
    """
    with patch(
        "socket.getaddrinfo",
        autospec=True,
        side_effect=socket.gaierror("[Errno 11004] getaddrinfo failed"),
    ):
        rtn = win_network.connect("test-server", 80)
        assert rtn
        assert not rtn["result"]
        assert (
            rtn["comment"]
            == "Unable to connect to test-server (unknown) on tcp port 80"
        )



import logging
import os

import pytest

import salt.serializers.json as jsonserializer
import salt.serializers.msgpack as msgpackserializer
import salt.serializers.plist as plistserializer
import salt.serializers.python as pythonserializer
import salt.serializers.yaml as yamlserializer
import salt.states.file as filestate
import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        filestate: {
            "__env__": "base",
            "__salt__": {"file.manage_file": False},
            "__serializers__": {
                "yaml.serialize": yamlserializer.serialize,
                "yaml.seserialize": yamlserializer.serialize,
                "python.serialize": pythonserializer.serialize,
                "json.serialize": jsonserializer.serialize,
                "plist.serialize": plistserializer.serialize,
                "msgpack.serialize": msgpackserializer.serialize,
            },
            "__opts__": {"test": False, "cachedir": ""},
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {},
        }
    }


@pytest.mark.skip_on_windows(reason="Do not run on Windows")
def test_hardlink(tmp_path):
    """
    Test to create a hardlink.
    """

    name = str(tmp_path / "testfile.txt")
    target = str(tmp_path / "target.txt")
    with salt.utils.files.fopen(target, "w") as fp:
        fp.write("")

    test_dir = str(tmp_path)
    user, group = "salt", "saltstack"

    def return_val(**kwargs):
        res = {
            "name": name,
            "result": False,
            "comment": "",
            "changes": {},
        }
        res.update(kwargs)
        return res

    mock_t = MagicMock(return_value=True)
    mock_f = MagicMock(return_value=False)
    mock_empty = MagicMock(return_value="")
    mock_uid = MagicMock(return_value="U1001")
    mock_gid = MagicMock(return_value="g1001")
    mock_nothing = MagicMock(return_value={})
    mock_stats = MagicMock(return_value={"inode": 1})
    mock_execerror = MagicMock(side_effect=CommandExecutionError)

    patches = {}
    patches["file.user_to_uid"] = mock_empty
    patches["file.group_to_gid"] = mock_empty
    patches["user.info"] = mock_empty
    patches["file.is_hardlink"] = mock_t
    patches["file.stats"] = mock_empty

    # Argument validation
    with patch.dict(filestate.__salt__, patches):
        expected = "Must provide name to file.hardlink"
        ret = return_val(comment=expected, name="")
        assert filestate.hardlink("", target) == ret

    # User validation for dir_mode
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_empty}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.object(
        os.path, "isabs", mock_t
    ):
        expected = f"User {user} does not exist"
        ret = return_val(comment=expected, name=name)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Group validation for dir_mode
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_empty}), patch.object(
        os.path, "isabs", mock_t
    ):
        expected = f"Group {group} does not exist"
        ret = return_val(comment=expected, name=name)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Absolute path for name
    nonabs = "./non-existent-path/to/non-existent-file"
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}):
        expected = f"Specified file {nonabs} is not an absolute path"
        ret = return_val(comment=expected, name=nonabs)
        assert filestate.hardlink(nonabs, target, user=user, group=group) == ret

    # Absolute path for target
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}):
        expected = f"Specified target {nonabs} is not an absolute path"
        ret = return_val(comment=expected, name=name)
        assert filestate.hardlink(name, nonabs, user=user, group=group) == ret
    # Test option -- nonexistent target
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.object(
        os.path, "exists", mock_f
    ), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        expected = f"Target {target} for hard link does not exist"
        ret = return_val(comment=expected, name=name)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Test option -- target is a directory
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.object(
        os.path, "exists", mock_t
    ), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        expected = f"Unable to hard link from directory {test_dir}"
        ret = return_val(comment=expected, name=name)
        assert filestate.hardlink(name, test_dir, user=user, group=group) == ret

    # Test option -- name is a directory
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        expected = f"Unable to hard link to directory {test_dir}"
        ret = return_val(comment=expected, name=test_dir)
        assert filestate.hardlink(test_dir, target, user=user, group=group) == ret

    # Test option -- name does not exist
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        expected = f"Hard link {name} to {target} is set for creation"
        changes = dict(new=name)
        ret = return_val(result=None, comment=expected, name=name, changes=changes)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Test option -- hardlink matches
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_t}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_stats}
    ), patch.object(
        os.path, "exists", mock_t
    ), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        expected = f"The hard link {name} is presently targetting {target}"
        ret = return_val(result=True, comment=expected, name=name)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Test option -- hardlink does not match
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_t}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_nothing}
    ), patch.object(
        os.path, "exists", mock_t
    ), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        expected = f"Link {name} target is set to be changed to {target}"
        changes = dict(change=name)
        ret = return_val(result=None, comment=expected, name=name, changes=changes)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Test option -- force removal
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_f}
    ), patch.object(
        os.path, "exists", mock_t
    ), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        expected = (
            "The file or directory {} is set for removal to "
            "make way for a new hard link targeting {}".format(name, target)
        )
        ret = return_val(result=None, comment=expected, name=name)
        assert (
            filestate.hardlink(name, target, force=True, user=user, group=group) == ret
        )

    # Test option -- without force removal
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_f}
    ), patch.object(
        os.path, "exists", mock_t
    ), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        expected = (
            "File or directory exists where the hard link {} "
            "should be. Did you mean to use force?".format(name)
        )
        ret = return_val(result=False, comment=expected, name=name)
        assert (
            filestate.hardlink(name, target, force=False, user=user, group=group) == ret
        )

    # Target is a directory
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}):
        expected = f"Unable to hard link from directory {test_dir}"
        ret = return_val(comment=expected, name=name)
        assert filestate.hardlink(name, test_dir, user=user, group=group) == ret

    # Name is a directory
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}):
        expected = f"Unable to hard link to directory {test_dir}"
        ret = return_val(comment=expected, name=test_dir)
        assert filestate.hardlink(test_dir, target, user=user, group=group) == ret

    # Try overwrite file with link
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_f}
    ), patch.object(
        os.path, "isfile", mock_t
    ):

        expected = f"File exists where the hard link {name} should be"
        ret = return_val(comment=expected, name=name)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Try overwrite link with same
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_t}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_stats}
    ), patch.object(
        os.path, "isfile", mock_f
    ):

        expected = "Target of hard link {} is already pointing to {}".format(
            name, target
        )
        ret = return_val(result=True, comment=expected, name=name)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Really overwrite link with same
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_t}
    ), patch.dict(
        filestate.__salt__, {"file.link": mock_t}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_nothing}
    ), patch.object(
        os, "remove", mock_t
    ), patch.object(
        os.path, "isfile", mock_f
    ):

        expected = f"Set target of hard link {name} -> {target}"
        changes = dict(new=name)
        ret = return_val(result=True, comment=expected, name=name, changes=changes)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Fail at overwriting link with same
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_t}
    ), patch.dict(
        filestate.__salt__, {"file.link": mock_execerror}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_nothing}
    ), patch.object(
        os, "remove", mock_t
    ), patch.object(
        os.path, "isfile", mock_f
    ):

        expected = "Unable to set target of hard link {} -> {}: {}".format(
            name, target, ""
        )
        ret = return_val(result=False, comment=expected, name=name)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Make new link
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_f}
    ), patch.dict(
        filestate.__salt__, {"file.link": mock_f}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_nothing}
    ), patch.object(
        os, "remove", mock_t
    ), patch.object(
        os.path, "isfile", mock_f
    ):

        expected = f"Created new hard link {name} -> {target}"
        changes = dict(new=name)
        ret = return_val(result=True, comment=expected, name=name, changes=changes)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Fail while making new link
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_f}
    ), patch.dict(
        filestate.__salt__, {"file.link": mock_execerror}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_nothing}
    ), patch.object(
        os, "remove", mock_t
    ), patch.object(
        os.path, "isfile", mock_f
    ):

        expected = "Unable to create new hard link {} -> {}: {}".format(
            name, target, ""
        )
        ret = return_val(result=False, comment=expected, name=name)
        assert filestate.hardlink(name, target, user=user, group=group) == ret

    # Force making new link over file
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_f}
    ), patch.dict(
        filestate.__salt__, {"file.link": mock_t}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_nothing}
    ), patch.object(
        os, "remove", mock_t
    ), patch.object(
        os.path, "isfile", mock_t
    ):

        expected = f"Created new hard link {name} -> {target}"
        changes = dict(new=name)
        changes["forced"] = "File for hard link was forcibly replaced"
        ret = return_val(result=True, comment=expected, name=name, changes=changes)
        assert (
            filestate.hardlink(name, target, user=user, force=True, group=group) == ret
        )

    # Force making new link over file but error out
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_f}
    ), patch.dict(
        filestate.__salt__, {"file.link": mock_execerror}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_nothing}
    ), patch.object(
        os, "remove", mock_t
    ), patch.object(
        os.path, "isfile", mock_t
    ):

        expected = "Unable to create new hard link {} -> {}: {}".format(
            name, target, ""
        )
        changes = dict(forced="File for hard link was forcibly replaced")
        ret = return_val(result=False, comment=expected, name=name, changes=changes)
        assert (
            filestate.hardlink(name, target, user=user, force=True, group=group) == ret
        )

    patches = {}
    patches["file.user_to_uid"] = mock_empty
    patches["file.group_to_gid"] = mock_empty
    patches["file.is_hardlink"] = mock_t
    patches["file.stats"] = mock_empty

    # Make new link when group is None and file.gid_to_group is unavailable
    with patch.dict(filestate.__salt__, patches), patch.dict(
        filestate.__salt__, {"file.user_to_uid": mock_uid}
    ), patch.dict(filestate.__salt__, {"file.group_to_gid": mock_gid}), patch.dict(
        filestate.__salt__, {"file.is_hardlink": mock_f}
    ), patch.dict(
        filestate.__salt__, {"file.link": mock_f}
    ), patch.dict(
        filestate.__salt__, {"file.stats": mock_nothing}
    ), patch.object(
        os, "remove", mock_t
    ), patch.object(
        os.path, "isfile", mock_f
    ):

        group = None
        expected = f"Created new hard link {name} -> {target}"
        changes = dict(new=name)
        ret = return_val(result=True, comment=expected, name=name, changes=changes)
        assert filestate.hardlink(name, target, user=user, group=group) == ret



import copy
import logging
import multiprocessing
import random
import time
from contextlib import contextmanager

import pytest
from saltfactories.utils import random_string

import salt.transport.zeromq
import salt.utils.process
from tests.support.mock import MagicMock, patch
from tests.support.pytest.transport import PubServerChannelProcess

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
    pytest.mark.skip_on_freebsd(reason="Temporarily skipped on FreeBSD."),
    pytest.mark.skip_on_spawning_platform(
        reason="These tests are currently broken on spawning platforms. Need to be rewritten.",
    ),
    pytest.mark.skipif(
        "grains['osfinger'] == 'Rocky Linux-8' and grains['osarch'] == 'aarch64'",
        reason="Temporarily skip on Rocky Linux 8 Arm64",
    ),
]


class PubServerChannelSender:
    def __init__(self, pub_server_channel, payload_list):
        self.pub_server_channel = pub_server_channel
        self.payload_list = payload_list

    def run(self):
        for payload in self.payload_list:
            self.pub_server_channel.publish(payload)
        time.sleep(2)


def generate_msg_list(msg_cnt, minions_list, broadcast):
    msg_list = []
    for i in range(msg_cnt):
        for idx, minion_id in enumerate(minions_list):
            if broadcast:
                msg_list.append(
                    {"tgt_type": "grain", "tgt": "id:*", "jid": msg_cnt * idx + i}
                )
            else:
                msg_list.append(
                    {"tgt_type": "list", "tgt": [minion_id], "jid": msg_cnt * idx + i}
                )
    return msg_list


@contextmanager
def channel_publisher_manager(msg_list, p_cnt, pub_server_channel):
    process_list = []
    msg_list = copy.deepcopy(msg_list)
    random.shuffle(msg_list)
    batch_size = len(msg_list) // p_cnt
    list_batch = [
        [x * batch_size, x * batch_size + batch_size] for x in range(0, p_cnt)
    ]
    list_batch[-1][1] = list_batch[-1][1] + 1
    try:
        for i, j in list_batch:
            c = PubServerChannelSender(pub_server_channel, msg_list[i:j])
            p = salt.utils.process.Process(target=c.run)
            process_list.append(p)
        for p in process_list:
            p.start()
        yield
    finally:
        for p in process_list:
            p.join()


@pytest.mark.skip_on_windows
@pytest.mark.slow_test
def test_zeromq_filtering_minion(salt_master, salt_minion):
    opts = dict(
        salt_master.config.copy(),
        ipc_mode="ipc",
        pub_hwm=0,
        zmq_filtering=True,
        acceptance_wait_time=5,
    )
    minion_opts = dict(
        salt_minion.config.copy(),
        zmq_filtering=True,
    )
    messages = 200
    workers = 5
    minions = 3
    expect = set(range(messages))
    target_minion_id = salt_minion.id
    minions_list = [target_minion_id]
    for _ in range(minions - 1):
        minions_list.append(random_string("zeromq-minion-"))
    msg_list = generate_msg_list(messages, minions_list, False)
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(
            return_value={
                "minions": minions_list,
                "missing": [],
                "ssh_minions": False,
            }
        ),
    ):
        with PubServerChannelProcess(opts, minion_opts) as server_channel:
            with channel_publisher_manager(
                msg_list, workers, server_channel.pub_server_channel
            ):
                cnt = 0
                last_results_len = 0
                while cnt < 20:
                    time.sleep(2)
                    results_len = len(server_channel.collector.results)
                    if last_results_len == results_len:
                        break
                    last_results_len = results_len
                    cnt += 1
        results = set(server_channel.collector.results)
        assert (
            results == expect
        ), f"{len(results)}, != {len(expect)}, difference: {expect.difference(results)} {results}"


@pytest.mark.skip_on_windows
@pytest.mark.slow_test
def test_zeromq_filtering_syndic(salt_master, salt_minion):
    opts = dict(
        salt_master.config.copy(),
        ipc_mode="ipc",
        pub_hwm=0,
        zmq_filtering=True,
        acceptance_wait_time=5,
        order_masters=True,
    )
    minion_opts = dict(
        salt_minion.config.copy(),
        zmq_filtering=True,
        __role="syndic",
    )
    messages = 200
    workers = 5
    minions = 3
    expect = set(range(messages * minions))
    minions_list = []
    for _ in range(minions):
        minions_list.append(random_string("zeromq-minion-"))
    msg_list = generate_msg_list(messages, minions_list, False)
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(
            return_value={
                "minions": minions_list,
                "missing": [],
                "ssh_minions": False,
            }
        ),
    ):
        with PubServerChannelProcess(opts, minion_opts) as server_channel:
            with channel_publisher_manager(
                msg_list, workers, server_channel.pub_server_channel
            ):
                cnt = 0
                last_results_len = 0
                while cnt < 20:
                    time.sleep(2)
                    results_len = len(server_channel.collector.results)
                    if last_results_len == results_len:
                        break
                    last_results_len = results_len
                    cnt += 1
        results = set(server_channel.collector.results)
        assert (
            results == expect
        ), f"{len(results)}, != {len(expect)}, difference: {expect.difference(results)} {results}"


@pytest.mark.skip_on_windows
@pytest.mark.slow_test
def test_zeromq_filtering_broadcast(salt_master, salt_minion):
    """
    Test sending messages to publisher using UDP with zeromq_filtering enabled
    """
    opts = dict(
        salt_master.config.copy(),
        ipc_mode="ipc",
        pub_hwm=0,
        zmq_filtering=True,
        acceptance_wait_time=5,
    )
    minion_opts = dict(
        salt_minion.config.copy(),
        zmq_filtering=True,
    )
    messages = 200
    workers = 5
    minions = 3
    expect = set(range(messages * minions))
    target_minion_id = salt_minion.id
    minions_list = [target_minion_id]
    for _ in range(minions - 1):
        minions_list.append(random_string("zeromq-minion-"))
    msg_list = generate_msg_list(messages, minions_list, True)
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(
            return_value={
                "minions": minions_list,
                "missing": [],
                "ssh_minions": False,
            }
        ),
    ):
        with PubServerChannelProcess(opts, minion_opts) as server_channel:
            with channel_publisher_manager(
                msg_list, workers, server_channel.pub_server_channel
            ):
                cnt = 0
                last_results_len = 0
                while cnt < 20:
                    time.sleep(2)
                    results_len = len(server_channel.collector.results)
                    if last_results_len == results_len:
                        break
                    last_results_len = results_len
                    cnt += 1
        results = set(server_channel.collector.results)
        assert (
            results == expect
        ), f"{len(results)}, != {len(expect)}, difference: {expect.difference(results)} {results}"


def test_pub_channel(master_opts):
    server = salt.transport.zeromq.PublishServer(master_opts)

    manager = multiprocessing.Manager()

    payloads = manager.list()

    def publish_payload(payload):
        server.publish_payload(payload)
        payloads.append(payload)

    process = multiprocessing.Process(
        target=server.publish_daemon, args=(publish_payload,)
    )
    process.start()

    server.publish({"meh": "bah"})

    start = time.monotonic()
    try:
        while not payloads:
            time.sleep(0.3)
            if time.monotonic() - start > 30:
                assert False, "No message received after 30 seconds"
    finally:
        server.close()
        process.terminate()
        process.join()


def test_pub_channel_filtering(master_opts):
    master_opts["zmq_filtering"] = True
    server = salt.transport.zeromq.PublishServer(master_opts)

    manager = multiprocessing.Manager()

    payloads = manager.list()

    def publish_payload(payload):
        server.publish_payload(payload)
        payloads.append(payload)

    process = multiprocessing.Process(
        target=server.publish_daemon, args=(publish_payload,)
    )
    process.start()

    server.publish({"meh": "bah"})

    start = time.monotonic()
    try:
        while not payloads:
            time.sleep(0.3)
            if time.monotonic() - start > 30:
                assert False, "No message received after 30 seconds"
    finally:
        server.close()
        process.terminate()
        process.join()


def test_pub_channel_filtering_topic(master_opts):
    master_opts["zmq_filtering"] = True
    server = salt.transport.zeromq.PublishServer(master_opts)
    manager = multiprocessing.Manager()

    payloads = manager.list()

    def publish_payload(payload):
        server.publish_payload(payload, topic_list=["meh"])
        payloads.append(payload)

    process = multiprocessing.Process(
        target=server.publish_daemon, args=(publish_payload,)
    )
    process.start()

    server.publish({"meh": "bah"})

    start = time.monotonic()
    try:
        while not payloads:
            time.sleep(0.3)
            if time.monotonic() - start > 30:
                assert False, "No message received after 30 seconds"
    finally:
        server.close()
        process.terminate()
        process.join()
